from typing import TypeVar, Generic, Type

from tortoise import Model
from tortoise.contrib.pydantic import PydanticModel

from dry_core.services import Service
from dry_core.operations import operation


BaseModel = TypeVar("BaseModel", bound=Model)
PydanticBaseModel = TypeVar("PydanticBaseModel", bound=PydanticModel)


class TortoiseService(Service[BaseModel], Generic[BaseModel, PydanticBaseModel]):
    pydantic_model: Type[PydanticBaseModel]

    async def get_instance_as_pydantic(self) -> PydanticBaseModel:
        self.validate_instance_filled()
        return await self.pydantic_model.from_tortoise_orm(self.instance)

    @classmethod
    def _model_fields(cls) -> set[str]:
        return cls.model._meta.fields

    @operation
    async def create(self, **kwargs) -> BaseModel:
        self.instance = await self.model.create(**kwargs)
        return self.instance

    @operation
    async def update(self, **kwargs) -> BaseModel:
        self.validate_instance_filled()
        for attr, value in kwargs.items():
            if attr in self._model_fields():
                setattr(self.instance, attr, value)
        await self.instance.save()
        return self.instance

    @operation
    async def delete(self) -> BaseModel:
        self.validate_instance_filled()
        deleted_instance = self.instance
        await self.instance.delete()
        return deleted_instance



from typing import Any, Type, Optional, Callable, List
from pydantic import BaseModel


class FunctionArgInfo(BaseModel):
    name: str
    types: Optional[List[Type]] = None
    default_value: Optional[Any] = None

    def check_type(self, type: Type) -> bool:
        if self.types is None:
            return False
        return any([issubclass(t, type) for t in self.types])


class FunctionInformation(BaseModel):
    function_object: Callable
    priority: int = -1
    args: dict[str, FunctionArgInfo] = {}
    rollback: Optional[Callable] = None

    @property
    def name(self) -> str:
        return self.function_object.__name__

    @property
    def receives_all_kwargs(self) -> bool:
        return "kwargs" in self.args.keys()


class OperationExecutionInfo(BaseModel):
    exc: Optional[Exception] = None
    on_func: Optional[FunctionInformation] = None
    func_args: Optional[dict[str, Any]] = None

    class Config:
        arbitrary_types_allowed = (Exception,)



from typing import Any, Callable, Optional, Union, Dict

from pydantic import BaseModel


class Argument(BaseModel):
    default_value: Any
    from_name: Optional[str] = None
    from_initial_args: bool = False
    validators: Optional[list[Callable]] = None

    def __init__(self, _default_value, **kwargs):
        kwargs["default_value"] = _default_value
        super(Argument, self).__init__(**kwargs)


class ArgumentsStore:
    def __init__(self, operation_args_dict: dict[str, Any], *, immutable: bool = False):
        self._operation_args_dict = operation_args_dict.copy()
        self._immutable = immutable

    @property
    def immutable(self) -> bool:
        return self._immutable

    # Store dict-like interface definition

    def _validate_immutability(self) -> None:
        if self._immutable:
            raise ValueError("Store is immutable")

    def get(self, key):
        return self._operation_args_dict.get(key)

    def get_with_default(self, key, default):
        return self._operation_args_dict.get(key, default)

    def copy(self):
        return self._operation_args_dict.copy()

    def clear(self):
        self._validate_immutability()
        return self._operation_args_dict.clear()

    def has_arg(self, arg_name):
        return arg_name in self._operation_args_dict

    def keys(self):
        return self._operation_args_dict.keys()

    def items(self):
        return self._operation_args_dict.items()

    def dict(self) -> dict[str, Any]:
        return self._operation_args_dict.copy()

    def update(self, **kwargs):
        self._validate_immutability()
        self._operation_args_dict.update(**kwargs)

    def __getitem__(self, arg_name):
        return self._operation_args_dict[arg_name]

    def __setitem__(self, arg_name, value):
        self._validate_immutability()
        self._operation_args_dict[arg_name] = value
        return self._operation_args_dict[arg_name]

    def __delitem__(self, key):
        self._validate_immutability()
        del self._operation_args_dict[key]

    def __repr__(self):
        return repr(self._operation_args_dict)

    def __len__(self):
        return len(self._operation_args_dict)

    def __eq__(self, other_store: Union["ArgumentsStore", Dict[str, Any]]):
        return set(self.items()) == set(other_store.items())

    def __contains__(self, item):
        return item in self._operation_args_dict

    def __iter__(self):
        return iter(self._operation_args_dict)



from typing import Optional, Generic, Type, AsyncIterable, Iterable, Union, TypeVar

from httpx import Client as HttpClient, AsyncClient as HttpAsyncClient, URL

from dry_core.exceptions import NotFoundError
from dry_core.selectors.generics import BaseSelector
from .pagination import BasePaginationModel, PaginatedRequest


SelectorInstance = TypeVar("SelectorInstance")
SelectorListItemInstance = TypeVar("SelectorListItemInstance")


class BaseAPIMixin:
    base_url: Optional[Union[str, URL]] = None
    connection_timeout: Optional[float] = 10.0
    headers: dict[str, str] = {}

    @classmethod
    def client(cls) -> HttpClient:
        return HttpClient(
            base_url=cls.base_url,
            headers=cls.headers,
            timeout=cls.connection_timeout,
        )

    @classmethod
    def async_client(cls) -> HttpAsyncClient:
        return HttpAsyncClient(
            base_url=cls.base_url,
            headers=cls.headers,
            timeout=cls.connection_timeout,
        )


class BaseAPISelector(
    BaseAPIMixin, BaseSelector[SelectorInstance], Generic[SelectorInstance, SelectorListItemInstance]
):
    list_item_model: Optional[SelectorListItemInstance] = None
    list_api_endpoint: str
    list_api_endpoint_pagination_model: Optional[Type[BasePaginationModel]] = None
    entity_api_endpoint: str

    # utils
    @classmethod
    def get_list_item_model_class(cls) -> Union[SelectorInstance, SelectorListItemInstance]:
        return cls.list_item_model or cls.model

    # selecting method
    @classmethod
    def get(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> Optional[SelectorInstance]:
        with cls.client() as client:
            endpoint = cls.entity_api_endpoint.format(**kwargs)
            response = client.get(endpoint, params=query_params)
            if response.status_code != 200:
                if raise_exception:
                    raise NotFoundError(
                        f"Error when trying get object (HTTP status code: {response.status_code}, response: {response.text}). "
                        f"Url: {endpoint}, {query_params=}, {kwargs=}"
                    )
                else:
                    return None
            return cls.model(**response.json())

    @classmethod
    async def aget(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> Optional[SelectorInstance]:
        async with cls.async_client() as client:
            endpoint = cls.entity_api_endpoint.format(**kwargs)
            response = await client.get(endpoint, params=query_params)
            if response.status_code != 200:
                if raise_exception:
                    raise NotFoundError(f"Error when trying get object. Url: {endpoint}, {query_params=}, {kwargs=}")
                else:
                    return None
            return cls.model(**response.json())

    @classmethod
    def list_get_all_g(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> Iterable[SelectorInstance]:
        with cls.client() as client:
            endpoint = cls.list_api_endpoint.format(**kwargs)
            if cls.list_api_endpoint_pagination_model is None:
                response = client.get(endpoint, params=query_params)
                yield [cls.get_list_item_model_class()(**model_data) for model_data in response.json()]
            else:
                paginated_request = cls.list_api_endpoint_pagination_model.get_first_page_request_params(
                    PaginatedRequest(url=endpoint, query_params=query_params)
                )
                response = client.get(paginated_request.url, params=paginated_request.query_params)
                page = cls.list_api_endpoint_pagination_model(**response.json())
                yield page.get_page_results(model_class=cls.get_list_item_model_class())
                while page.have_next_page:
                    paginated_request = page.get_next_page_request_params(
                        request=PaginatedRequest(url=cls.list_api_endpoint, query_params=query_params)
                    )
                    if paginated_request is None:
                        return
                    response = client.get(paginated_request.url, params=paginated_request.query_params)
                    page = cls.list_api_endpoint_pagination_model(**response.json())
                    yield page.get_page_results(model_class=cls.get_list_item_model_class())

    @classmethod
    def list_get_all(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> Optional[list[SelectorInstance]]:
        results = []
        for part in cls.list_get_all_g(query_params=query_params, raise_exception=raise_exception, **kwargs):
            results += part
        return results

    @classmethod
    async def alist_get_all_g(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> AsyncIterable[list[SelectorInstance]]:
        async with cls.async_client() as client:
            endpoint = cls.list_api_endpoint.format(**kwargs)
            if cls.list_api_endpoint_pagination_model is None:
                response = await client.get(endpoint, params=query_params)
                yield [cls.get_list_item_model_class()(**model_data) for model_data in response.json()]
            else:
                paginated_request = cls.list_api_endpoint_pagination_model.get_first_page_request_params(
                    request=PaginatedRequest(url=endpoint, query_params=query_params)
                )
                response = await client.get(paginated_request.url, params=paginated_request.query_params)
                page = cls.list_api_endpoint_pagination_model(**response.json())
                yield page.get_page_results(model_class=cls.get_list_item_model_class())
                while page.have_next_page:
                    paginated_request = page.get_next_page_request_params(
                        request=PaginatedRequest(url=cls.list_api_endpoint, query_params=query_params),
                    )
                    if paginated_request is None:
                        return
                    response = await client.get(paginated_request.url, params=paginated_request.query_params)
                    page = cls.list_api_endpoint_pagination_model(**response.json())
                    yield page.get_page_results(model_class=cls.get_list_item_model_class())

    @classmethod
    async def alist_get_all(
        cls, *, query_params: Optional[dict[str, str]] = None, raise_exception: bool = False, **kwargs
    ) -> Optional[list[SelectorInstance]]:
        results = []
        async for part in cls.alist_get_all_g(query_params=query_params, raise_exception=raise_exception, **kwargs):
            results += part
        return results



from typing import Optional, Generic, TypeVar, Type, Union
from uuid import UUID

from dry_core.selectors.generics import BaseSelector

ServiceInstance = TypeVar("ServiceInstance")
BaseInstanceSelector = TypeVar("BaseInstanceSelector", bound=BaseSelector)


class BaseServiceMixin(Generic[ServiceInstance]):
    model: Type[ServiceInstance]

    def __init__(self, *args, **kwargs):
        self.validate_service_model_defined()

    def validate_service_model_defined(self) -> None:
        if self.model is None:
            raise AttributeError(f'Service class {self.__class__.__name__} must setup "model" field')


class ServiceInstanceMixin(BaseServiceMixin[ServiceInstance], Generic[ServiceInstance]):
    def __init__(self, instance: Optional[ServiceInstance] = None, *args, **kwargs):
        super(ServiceInstanceMixin, self).__init__(*args, **kwargs)
        self.validate_service_model_defined()
        # if instance don't set on higher level, like it can be in UUIDServiceMixin, try init it from arguments
        if getattr(self, "instance", None) is None:
            self.instance: Optional[ServiceInstance] = instance

    def validate_instance_filled(self, validate_type: bool = True, raise_exception: bool = True) -> bool:
        if self.instance is None:
            if raise_exception:
                raise ValueError(
                    '{class_name} object field "instance" is None'.format(class_name=self.__class__.__name__)
                )
            return False
        if validate_type:
            if not isinstance(self.instance, self.model):
                if raise_exception:
                    raise TypeError(
                        '{class_name} field "instance" is type {instance_type}, must be {service_instance_model}'.format(
                            class_name=self.__class__.__name__,
                            instance_type=self.instance.__class__.__name__,
                            service_instance_model=self.model.__name__,
                        )
                    )
                return False
        return True


class UUIDServiceInstanceMixin(ServiceInstanceMixin[ServiceInstance], Generic[ServiceInstance]):
    selector: Type[BaseInstanceSelector]
    selector_get_by_uuid_method_name: str = "get_by_uuid"
    selector_get_by_uuid_param_name: str = "uuid"

    def __init__(self, *args, uuid: Optional[Union[UUID, str]] = None, **kwargs):
        if self.selector is None:
            raise AttributeError(f'"selector" class attr must be defined for {self.__class__}')
        if getattr(self.selector, self.selector_get_by_uuid_method_name, None) is None:
            raise AttributeError(f'"selector" class must have method "{self.selector_get_by_uuid_method_name}"')
        if uuid is not None:
            if isinstance(uuid, UUID):
                uuid = str(uuid)
            self.instance: ServiceInstance = getattr(self.selector, self.selector_get_by_uuid_method_name)(
                **{self.selector_get_by_uuid_param_name: uuid}
            )
        super(UUIDServiceInstanceMixin, self).__init__(*args, **kwargs)



import functools
import inspect
from typing import Callable, Type, Any, Optional

from .arguments import ArgumentsStore, Argument
from ._models import FunctionInformation, FunctionArgInfo, OperationExecutionInfo
from ._utils import parse_function_information
from ..exceptions import OperationImproperlyConfigured


class _OperationBase:
    def __init__(self, method: Callable, method_class: Type):
        functools.update_wrapper(self, method)
        self.method: FunctionInformation = parse_function_information(method)
        self.method_class: Type = method_class
        self._result = None

    @property
    def operation_name(self) -> str:
        return self.method.name

    @property
    def result(self) -> Any:
        return self._result

    def register_result(self, value) -> Any:
        self._result = value
        return self._result

    def clear_result(self) -> None:
        self._result = None

    @property
    def operation_method(self) -> str:
        return f"{self.method_class.__name__}.{self.method.function_object.__name__}"

    def __call__(self, *args, **kwargs):
        return self.method.function_object(*args, **kwargs)


class Operation(_OperationBase):
    # Setup and clean up method

    def __init__(
        self,
        *args,
        exception_handlers_mapping: Optional[dict[Type, Callable]] = None,
        as_transaction: bool = False,
        rollback: Optional[Callable] = None,
        **kwargs,
    ):

        super(Operation, self).__init__(*args, **kwargs)

        self.method.rollback = rollback

        self._in_transaction_by_default = as_transaction
        self._exception_handlers_mapping = exception_handlers_mapping

        if self._in_transaction_by_default and self._exception_handlers_mapping is not None:
            raise OperationImproperlyConfigured(
                "You cant use both 'exception_handlers_mapping' and 'as_transaction=True' at one time."
            )

        # Attributes that changes on each operation call
        self._operation_self_obj = None
        self.initial_arguments_store: Optional[ArgumentsStore] = None
        self.arguments_store: Optional[ArgumentsStore] = None
        self._executed_functions: list[FunctionInformation] = []
        self._execution_info = OperationExecutionInfo()

    @property
    def executed_functions(self) -> list[FunctionInformation]:
        return self._executed_functions

    def _setup_operation_state(self, *args, **kwargs):
        try:
            self._setup_operation_self_obj(args[0])
        except IndexError:
            raise ValueError(f'Cant extract "self" for method {self.operation_method} on preparing of operation')
        self._load_pre_and_post_operation_func()
        self._setup_operation_args_stores(args[1:], **kwargs)
        self._setup_operation_transaction_mode()
        self._execution_info = OperationExecutionInfo()
        self._executed_functions: list[FunctionInformation] = []

    def _clean_up_operation_state(self):
        self.clear_result()
        self._executed_functions: list[FunctionInformation] = []
        self._execution_info = OperationExecutionInfo()
        self._clean_up_operation_transaction_mode()
        self._clean_up_operation_args_stores()
        self._clean_up_operation_self_obj()

    def _setup_operation_self_obj(self, operation_self_obj) -> None:
        self._operation_self_obj = operation_self_obj

    def _clean_up_operation_self_obj(self) -> None:
        self._operation_self_obj = None

    def _setup_operation_args_stores(self, *args, **kwargs):
        self.initial_arguments_store = ArgumentsStore(operation_args_dict=kwargs, immutable=True)
        self.arguments_store = ArgumentsStore(operation_args_dict=kwargs)

    def _clean_up_operation_args_stores(self):
        self.initial_arguments_store = None
        self.arguments_store = None

    def _setup_operation_transaction_mode(self):
        in_transaction = self.initial_arguments_store.get_with_default(
            "operation_as_transaction", self._in_transaction_by_default
        )
        if not in_transaction and self._in_transaction_by_default:
            raise OperationImproperlyConfigured(
                "Operation transaction mechanism cant be disabled if it have been enabled on operation definition level"
            )
        self._transactions_manager = OperationTransactionManager(
            in_transaction, exception_handlers_mapping=self._exception_handlers_mapping
        )

    def _clean_up_operation_transaction_mode(self):
        self._transactions_manager = None

    def _load_pre_and_post_operation_func(self):
        from .decorators import pre, post

        self._pre_operation_funcs: list[FunctionInformation] = pre.get_additional_operation_funcs(
            operation_owner_class=self.method_class, operation_name=self.operation_name
        )
        self._post_operation_funcs: list[FunctionInformation] = post.get_additional_operation_funcs(
            operation_owner_class=self.method_class, operation_name=self.operation_name
        )

    # Core logic
    def __call__(self, *args, **kwargs):
        try:
            return self._main_pipeline(*args, **kwargs)
        except Exception as exc:
            if isinstance(exc, OperationImproperlyConfigured):
                raise exc
            if self._transactions_manager.in_transaction:
                self._transactions_manager.rollback_executed_functions(self)
            else:
                if self._transactions_manager.exception_handlers_mapping is None:
                    raise exc
                # Try to handle exception if transaction mechanism not enabled & have exception handlers
                try:
                    result = self._transactions_manager.handle_exception(exc, self)
                    self._clean_up_operation_state()
                except Exception as e:
                    self._clean_up_operation_state()
                    raise e
                return result

    def _main_pipeline(self, *args, **kwargs) -> Any:
        self._setup_operation_state(*args, **kwargs)

        for additional_func in self._pre_operation_funcs:
            self._execute_function(additional_func)
        self.register_result(self._execute_function(self.method))
        for additional_func in self._post_operation_funcs:
            self._execute_function(additional_func)

        result = self.result
        self._clean_up_operation_state()
        return result

    def _execute_function(self, func_info: FunctionInformation) -> Any:
        self._execution_info.on_func = func_info
        args = self._get_args_for_func(func_info)
        self._execution_info.func_args = args.copy()
        result = func_info.function_object(self._operation_self_obj, **args)
        self._executed_functions.append(func_info)
        self._execution_info.on_func = None
        self._execution_info.func_args = None
        return result

    def _get_arg_value_for_func_by_arg_info(self, arg_info: FunctionArgInfo) -> Any:
        if arg_info.types is not None and arg_info.check_type(Operation):
            return self
        elif not isinstance(arg_info.default_value, Argument):
            value = (
                self.arguments_store[arg_info.name] if arg_info.name in self.arguments_store else arg_info.default_value
            )
        else:
            argument_metadata_obj: Argument = arg_info.default_value
            args_store_to_use: ArgumentsStore = (
                self.arguments_store if not argument_metadata_obj.from_initial_args else self.initial_arguments_store
            )
            arg_name_in_store = argument_metadata_obj.from_name or arg_info.name
            value = (
                args_store_to_use[arg_name_in_store]
                if arg_name_in_store in args_store_to_use
                else argument_metadata_obj.default_value
            )
            # TODO add type validation for value
            if argument_metadata_obj.validators is not None:
                for validator in argument_metadata_obj.validators:
                    value = validator(value)

        return value

    def _get_args_for_func(self, func_info: FunctionInformation) -> dict[str, Any]:
        args = {}
        for arg_name, arg_info in func_info.args.items():
            if arg_name in ["self", "kwargs"]:
                continue
            value = self._get_arg_value_for_func_by_arg_info(arg_info=arg_info)
            if value is Ellipsis:
                continue
            args[arg_name] = value
        if func_info.receives_all_kwargs:
            # add all arguments that not been already added
            for arg_name, value in self.arguments_store.items():
                if arg_name in ["self", "kwargs"]:
                    continue
                args[arg_name] = value
        return args


class _AsyncOperation(Operation):
    async def __call__(self, *args, **kwargs):
        try:
            return await self._main_pipeline(*args, **kwargs)
        except Exception as exc:
            if isinstance(exc, OperationImproperlyConfigured):
                raise exc
            if self._transactions_manager.in_transaction:
                await self._transactions_manager.arollback_executed_functions(self)
            else:
                if self._transactions_manager.exception_handlers_mapping is None:
                    raise exc
                # Try to handle exception if transaction mechanism not enabled & have exception handlers
                try:
                    result = await self._transactions_manager.ahandle_exception(exc, self)
                    self._clean_up_operation_state()
                except Exception as e:
                    self._clean_up_operation_state()
                    raise e
                return result

    async def _main_pipeline(self, *args, **kwargs) -> Any:
        self._setup_operation_state(*args, **kwargs)

        for additional_func in self._pre_operation_funcs:
            await self._execute_function(additional_func)
        result = await self._execute_function(self.method)
        for additional_func in self._post_operation_funcs:
            await self._execute_function(additional_func)

        self._clean_up_operation_state()
        return result

    async def _execute_function(self, func_info: FunctionInformation) -> Any:
        self._execution_info.on_func = func_info
        args = self._get_args_for_func(func_info)
        self._execution_info.func_args = args.copy()
        if inspect.iscoroutinefunction(func_info.function_object):
            result = await func_info.function_object(self._operation_self_obj, **args)
        else:
            result = func_info.function_object(self._operation_self_obj, **args)
        self._executed_functions.append(func_info)
        self._execution_info.func_args = None
        self._execution_info.on_func = None
        return result


class OperationTransactionManager:
    def __init__(
        self,
        in_transaction: bool,
        exception_handlers_mapping: Optional[dict[Type, Callable]] = None,
    ):
        self.in_transaction = in_transaction
        self.exception_handlers_mapping = exception_handlers_mapping

    def _get_handler_by_exception(self, exc: Exception) -> Optional[Callable]:
        for cls in inspect.getmro(exc.__class__):
            handler = self.exception_handlers_mapping.get(cls, None)
            if handler is not None:
                return handler

    def handle_exception(self, exc: Exception, op: Operation) -> Any:
        if self.exception_handlers_mapping is None:
            raise OperationImproperlyConfigured("'exception_handlers_mapping' is not configured")

        handler = self._get_handler_by_exception(exc)
        if handler is None:
            return exc
        else:
            return handler(exc, op)

    async def ahandle_exception(self, exc: Exception, op: Operation):
        if self.exception_handlers_mapping is None:
            raise OperationImproperlyConfigured("'exception_handlers_mapping' is not configured")

        handler = self._get_handler_by_exception(exc)
        if handler is None:
            return exc
        else:
            if inspect.iscoroutinefunction(handler):
                return await handler(exc, op)
            return handler(exc, op)

    def rollback_executed_functions(self, op: Operation):
        for func in op.executed_functions:
            if func.rollback is not None:
                func.rollback(op)

    async def arollback_executed_functions(self, op: Operation):
        for func in op.executed_functions:
            if func.rollback is not None:
                if inspect.iscoroutinefunction(func.rollback):
                    await func.rollback(op)
                else:
                    func.rollback(op)



from typing import Any, Optional


class DryPyException(Exception):
    def __init__(self, *args, additional_data: Optional[dict[str, Any]] = None):
        super(DryPyException, self).__init__(*args)
        self.additional_data: dict[str, Any] = additional_data or {}


class ApplicationError(DryPyException):
    pass


class NotFoundError(DryPyException):
    pass


# Operations Errors
class BaseOperationException(DryPyException):
    pass


class OperationImproperlyConfigured(BaseOperationException):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dry_core',
 'dry_core.exceptions',
 'dry_core.operations',
 'dry_core.selectors',
 'dry_core.selectors.api',
 'dry_core.selectors.tortoise',
 'dry_core.services',
 'dry_core.services.tortoise',
 'dry_core.utils']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.22,<0.23', 'pydantic>=1.9.0,<2.0.0']

extras_require = \
{'tortoise': ['tortoise-orm[asyncpg,accel]>=0.19.0,<0.20.0']}

setup_kwargs = {
    'name': 'dry-core',
    'version': '0.4.13',
    'description': '',
    'long_description': '# Dry-core\n\n`dry-core` is core package of `dry-*` package series. Main goal \nis to minimize and power up code, make it clear and easy supportable.\n\nDocumentation will be available soon.\n',
    'author': 'Илья Маркевич',
    'author_email': 'samuray21x@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Generic, TypeVar, Type, Optional

import tortoise
from tortoise.contrib.pydantic import PydanticModel
from tortoise.models import Model, QuerySetSingle, QuerySet

from dry_core.selectors.generics import BaseSelector

BaseModel = TypeVar("BaseModel", bound=Model)
PydanticBaseModel = TypeVar("PydanticBaseModel", bound=PydanticModel)


class TortoiseSelector(BaseSelector[BaseModel], Generic[BaseModel, PydanticBaseModel]):
    model: Type[BaseModel]
    pydantic_model: Type[PydanticBaseModel]
    id_field = "id"

    @classmethod
    async def get(cls, *args, **kwargs) -> Optional[BaseModel]:
        query_set = cls.get_as_qs(*args, **kwargs)
        if query_set is None:
            return None
        try:
            return await query_set
        except tortoise.exceptions.DoesNotExist:
            return None

    @classmethod
    def get_as_qs(cls, *args, **kwargs) -> QuerySetSingle[BaseModel]:
        try:
            id_value = args[0]
        except IndexError:
            id_value = kwargs.get(cls.id_field, None)
        if id_value is None:
            raise ValueError(f"id not provided. id must be first arg or kwarg with name '{cls.id_field}'")
        return cls.model.get(**{cls.id_field: id_value})

    @classmethod
    async def get_as_pydantic(cls, *args, **kwargs) -> Optional[PydanticBaseModel]:
        query_set = cls.get_as_qs(*args, **kwargs)
        if query_set is None:
            return None
        try:
            return await cls.convert_to_pydantic_from_query_set_single(query_set)
        except tortoise.exceptions.DoesNotExist:
            return None

    @classmethod
    async def list_get(cls, **kwargs) -> list[BaseModel]:
        return await cls.list_get_as_qs(**kwargs)

    @classmethod
    def list_get_as_qs(cls, **kwargs) -> QuerySet[BaseModel]:
        return cls.model.filter(**kwargs)

    @classmethod
    async def list_get_as_pydantic(cls, **kwargs) -> list[PydanticBaseModel]:
        query_set = cls.list_get_as_qs(**kwargs)
        return await cls.convert_to_pydantic_from_query_set(query_set)

    @classmethod
    async def list_get_all(cls) -> list[BaseModel]:
        return await cls.list_get_all_as_qs()

    @classmethod
    def list_get_all_as_qs(cls) -> QuerySet[BaseModel]:
        return cls.model.all()

    @classmethod
    async def list_get_all_as_pydantic(cls) -> list[PydanticBaseModel]:
        query_set = cls.list_get_all_as_qs()
        return await cls.convert_to_pydantic_from_query_set(query_set)

    # Utils

    @classmethod
    async def convert_to_pydantic_from_query_set(cls, query_set: QuerySet[BaseModel]) -> list[PydanticBaseModel]:
        cls._validate_pydantic_model()
        return await cls.pydantic_model.from_queryset(query_set)

    @classmethod
    async def convert_to_pydantic_from_query_set_single(cls, query_set: QuerySetSingle[BaseModel]) -> PydanticBaseModel:
        cls._validate_pydantic_model()
        return await cls.pydantic_model.from_queryset_single(query_set)

    @classmethod
    def _validate_pydantic_model(cls) -> None:
        if cls.pydantic_model is None:
            raise ValueError("'pydantic_model' field must be set")



from typing import Any, Optional, Type

from pydantic import BaseModel as PydanticBaseModel


class PaginatedRequest(PydanticBaseModel):
    url: str
    query_params: Optional[dict[str, str]] = None
    body_params: Optional[dict[str, Any]] = None


class BasePaginationModel(PydanticBaseModel):
    request: PaginatedRequest

    @property
    def have_next_page(self) -> bool:
        raise NotImplementedError

    @property
    def have_prev_page(self) -> bool:
        raise NotImplementedError

    @classmethod
    def get_first_page_request_params(cls, request: PaginatedRequest) -> PaginatedRequest:
        raise NotImplementedError

    def get_next_page_request_params(self, request: Optional[PaginatedRequest]) -> Optional[PaginatedRequest]:
        raise NotImplementedError

    def get_prev_page_request_params(self, request: Optional[PaginatedRequest]) -> Optional[PaginatedRequest]:
        raise NotImplementedError

    def get_page_results(self, model_class: Type) -> list[Any]:
        raise NotImplementedError


class LimitOffsetPaginationModel(BasePaginationModel):
    _default_limit: int = 50

    limit: int = _default_limit
    offset: int = 0
    count: int = 0
    next: Optional[str] = None
    previous: Optional[str] = None
    results: list[Any] = []

    @property
    def results_count(self) -> int:
        return len(self.results)

    @property
    def have_next_page(self) -> bool:
        return self.next is not None or self.offset + self.limit < self.count

    @property
    def have_prev_page(self) -> bool:
        return self.previous is not None or (self.offset > 0 and self.count > 0)

    @classmethod
    def get_first_page_request_params(cls, request: PaginatedRequest) -> PaginatedRequest:
        query_params = request.query_params or {}
        return PaginatedRequest(
            url=request.url,
            query_params=query_params
            | {
                "limit": query_params.get("limit", None) or cls._default_limit,
                "offset": query_params.get("offset", None) or 0,
            },
        )

    def get_next_page_request_params(self, request: Optional[PaginatedRequest] = None) -> Optional[PaginatedRequest]:
        if not self.have_next_page:
            return
        if self.next is not None:
            return PaginatedRequest(
                url=self.next,
                query_params=request.query_params if request is not None else None,
                body_params=request.body_params if request is not None else None,
            )
        if request is not None:
            query_params = request.query_params
        else:
            new_offset = self.offset + self.limit
            query_params = {
                "limit": self.count - new_offset if (new_offset + self.limit) >= self.count else self.limit,
                "offset": new_offset,
            }
        return PaginatedRequest(
            url=(request or self.request).url,
            query_params=query_params,
            body_params=request.body_params if request is not None else None,
        )

    def get_prev_page_request_params(self, request: Optional[PaginatedRequest] = None) -> Optional[PaginatedRequest]:
        if not self.have_prev_page:
            return
        if self.previous is not None:
            return PaginatedRequest(
                url=self.previous,
                query_params=request.query_params if request is not None else None,
                body_params=request.body_params if request is not None else None,
            )
        if request is not None:
            query_params = request.query_params
        else:
            new_offset = self.offset - self.limit
            query_params = {
                "limit": self.offset if new_offset > 0 else self.limit,
                "offset": new_offset if new_offset > 0 else 0,
            }
        return PaginatedRequest(
            url=(request or self.request).url,
            query_params=query_params,
            body_params=request.body_params if request is not None else None,
        )

    def get_page_results(self, model_class: Optional[Type] = None) -> list[Any]:
        if model_class is None:
            return self.results
        return [
            model_class(**instance_data) if isinstance(instance_data, dict) else model_class(instance_data)
            for instance_data in self.results
        ]



import functools
import inspect
from typing import Callable, Type, Iterable, TYPE_CHECKING, Optional

from ._models import FunctionInformation
from ._utils import parse_function_information


class _BaseOperationAdditionalFunctionDecorator:  # noqa
    """
    Base decorator class, that contain all common logic for
    store and work with decorated functions of services classes
    """

    def __init__(self):
        self._operations_additional_funcs_store: dict[Type, dict[str, list[FunctionInformation]]] = {}

    def __getattr__(self, operation_name):
        """
        Use getattr to dynamic registration of any operation additions
        """

        class _BaseOperationAdditionalFunctionDecoratorWrapper:
            def __init__(self, method=None, *, priority: int = -1, rollback: Optional[Callable] = None):
                self.method = method
                self.priority = priority
                self.rollback = rollback

                if method is not None:
                    functools.update_wrapper(self, method)

            def __set_name__(local_self, owner, name):
                nonlocal self

                self._register_operation_func(
                    operation_name=operation_name,
                    original_function_owner=owner,
                    original_function=local_self.method,
                    priority=local_self.priority,
                    rollback=local_self.rollback,
                )

                setattr(owner, name, local_self.method)

            def __call__(self, method):
                return _BaseOperationAdditionalFunctionDecoratorWrapper(
                    method, priority=self.priority, rollback=self.rollback
                )

        return _BaseOperationAdditionalFunctionDecoratorWrapper

    def _get_class_funcs(self, class_obj) -> dict[str, list[FunctionInformation]]:
        """
        Used in local cases and can be used only for add/edit class funcs,
        but can't handle MRO recursive funcs handling
        """
        if (res := self._operations_additional_funcs_store.get(class_obj, None)) is None:
            res = self._operations_additional_funcs_store[class_obj] = {}
        return res

    def _register_operation_func(
        self,
        operation_name: str,
        original_function_owner: Type,
        original_function: Callable,
        priority: int,
        rollback: Optional[Callable] = None,
    ):
        operation_name = operation_name.lower()
        funcs_store = self._get_class_funcs(original_function_owner)
        if operation_name not in funcs_store:
            funcs_store[operation_name] = []
        funcs_list = funcs_store[operation_name]

        func_info = parse_function_information(original_function)
        func_info.priority = priority
        func_info.rollback = rollback
        funcs_list.append(func_info)

    def get_additional_operation_funcs(self, operation_owner_class, operation_name: str) -> list[FunctionInformation]:
        classes: Iterable[Type] = reversed(inspect.getmro(operation_owner_class))
        result: list[FunctionInformation] = []
        for obj in classes:
            result.extend(self._get_class_funcs(obj).get(operation_name, []))
        return result


# Decorators collection for automatic running pre-operation functions
pre = _BaseOperationAdditionalFunctionDecorator()

# Decorators collection for automatic running post-operation functions
post = _BaseOperationAdditionalFunctionDecorator()


if TYPE_CHECKING:
    # defined here to help IDEs and static type checkers only
    def operation(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

else:

    class operation:
        def __init__(
            self,
            method=None,
            *,
            exception_handlers_mapping: Optional[dict[Type, Callable]] = None,
            as_transaction: bool = False,
            rollback: Optional[Callable] = None,
        ):
            self.method = method
            self.operation = None

            self.exception_handlers_mapping = exception_handlers_mapping
            self.as_transaction = as_transaction
            self.rollback = rollback

        def __set_name__(self, owner, name):
            from .operations import Operation, _AsyncOperation

            if inspect.iscoroutinefunction(self.method):
                self.operation = _AsyncOperation(
                    method=self.method,
                    method_class=owner,
                    exception_handlers_mapping=self.exception_handlers_mapping,
                    as_transaction=self.as_transaction,
                    rollback=self.rollback,
                )
            else:
                self.operation = Operation(
                    method=self.method,
                    method_class=owner,
                    exception_handlers_mapping=self.exception_handlers_mapping,
                    as_transaction=self.as_transaction,
                    rollback=self.rollback,
                )

            @functools.wraps(self.method)
            def wrapper(*args, **kwargs):
                return self.operation(*args, **kwargs)

            setattr(owner, name, wrapper)

        def __call__(self, func):
            return operation(
                func,
                exception_handlers_mapping=self.exception_handlers_mapping,
                as_transaction=self.as_transaction,
                rollback=self.rollback,
            )



from typing import Generic, TypeVar

from . import mixins

ServiceInstance = TypeVar("ServiceInstance")


class Service(mixins.ServiceInstanceMixin[ServiceInstance], Generic[ServiceInstance]):
    pass


class UUIDService(mixins.UUIDServiceInstanceMixin[ServiceInstance], Service[ServiceInstance], Generic[ServiceInstance]):
    pass



import inspect
from typing import Callable, get_args, Union, get_origin

from ._models import FunctionInformation, FunctionArgInfo


def parse_function_information(
    function: Callable,
) -> FunctionInformation:
    return FunctionInformation(
        function_object=function,
        args=parse_function_parameters(function),
    )


def parse_function_parameters(
    function: Callable,
) -> dict[str, FunctionArgInfo]:
    def _parse_types_by_parameter(parameter):
        # todo rewrite for recursive parsing
        annotation = parameter.annotation
        # if no annotation provided
        if annotation is inspect.Parameter.empty:
            return None

        # if Union type annotation convert to Union args
        # else convert to list of 1 element
        if get_origin(annotation) is Union:
            annotation = get_args(annotation)
        else:
            annotation = [annotation]

        return annotation

    return {
        parameter.name: FunctionArgInfo(
            name=parameter.name,
            types=_parse_types_by_parameter(parameter),
            default_value=parameter.default if parameter.default is not inspect.Parameter.empty else ...,
        )
        for _, parameter in inspect.signature(function).parameters.items()
    }
